unit LGP;

interface

uses SysUtils, Classes, Math, TE, SD;

type TLGP = class
        private
                individualData: array of TTuringEvaluator;

                evolutionTotal: array of double;
                bestEvolution : double;

                sample : TSampleData;

                crossoverRate : double;
                crossoverPop : integer;
                incrementalCount : integer;

                procedure Select(index: integer; value : double);
                function getEvolution(index : integer):double;
                procedure Crossover(indexFrom, indexTo : integer);
                procedure Mutation(index: integer);
                procedure Sort();

        public
                chromosomeCount, populationCount, chromosomeMax: integer;
                generationCount, bestClones: integer;

                incrementalSearch : boolean;

                constructor Create(chromoCount, popCount, chromoMax: integer);
                destructor Destroy();

                procedure Populate();
                procedure Evolve();
                procedure Evaluate();

                procedure setCrossoverRate(rate: double);
                function getCrossoverRate():double;
                function getIndividual(index: integer):TTuringEvaluator;

                procedure setSample(s : TSampleData);
                function getSample():TSampleData;

end;

implementation

        constructor TLGP.Create(chromoCount, popCount, chromoMax: integer);
        var i : integer;
        begin

                chromosomeCount := chromoCount;
                populationCount := popCount;
                chromosomeMax := chromoMax;
                bestEvolution := 99999999999;
                incrementalSearch := true;
                incrementalCount := 0;

                sample := TSampleData.Create(0);

                setCrossoverRate(0.20);    // s faz crossover em 20% da populao

                SetLength(individualData, popCount);
                SetLength(evolutionTotal, popCount);

                for i := 0 to populationCount - 1 do
                        individualData[i] := TTuringEvaluator.Create(chromosomeCount, chromosomeMax);

                Populate;
        end;

        destructor TLGP.Destroy();
        var i : integer;
        begin
                for i := 0 to populationCount - 1 do
                        if individualData[i] <> nil then
                                individualData[i].Free;
                inherited;
        end;

        procedure TLGP.Populate();
        var i : integer;
        begin
                generationCount := 0;

                for i := 0 to populationCount - 1 do
                begin
                        individualData[i].randomPopulate;
                        Select(i, 0);
                end;
                bestEvolution := 9999999999;

                if incrementalSearch then
                        incrementalCount := trunc(sample.rowCount * 0.10);  // start at 10% of sample
        end;

        procedure TLGP.Evaluate();
        var i, t, rowCount : integer;
            value, ls, dif : double;
            te : TTuringEvaluator;
        begin
                if incrementalSearch then
                        rowCount := incrementalCount - 1
                else
                        rowCount := sample.rowCount - 1;

                for i := 0 to populationCount - 1 do
                begin
                        te := getIndividual(i);
                        ls := 0;
                        for t := 0 to rowCount do
                        begin
                                value := te.Evaluate(sample.getX(t));

                                // least squares method
                                dif := (value - sample.getY(t));
                                ls := ls + sqrt(dif*dif);

                                if ls > bestEvolution then
                                        break;
                        end;
                        Select(i, ls);
                end;
        end;

        procedure TLGP.Evolve();
        var i : integer;
        begin
                generationCount := generationCount + 1;

                Sort;

                bestEvolution := getEvolution(0);
                bestClones := 0;
                Select(0, 0);

                if(incrementalSearch) then
                begin
                        if(bestEvolution < 0.0001) then
                        begin
                                incrementalCount := incrementalCount + trunc(sample.rowCount * 0.10);  // add 10% of sample
                                if incrementalCount > sample.rowCount then
                                        incrementalCount := sample.rowCount;
                        end;

                end;

                for i := 1 to populationCount - 1 do
                begin
                        if i <= crossoverPop then
                            Crossover(0, i)
                        else
                            Mutation(i);

                        if getEvolution(i) <= (bestEvolution*1.3) then
                            bestClones := bestClones + 1;
                        Select(i, 0);
                end;
        end;

        procedure TLGP.Sort();
        var i, k : integer;
            t : double;
            e : TTuringEvaluator;
        begin
            for i := 0 to populationCount - 2 do
            begin
                for k := i+1 to populationCount - 1 do
                begin
                    if evolutionTotal[k] < evolutionTotal[i] then
                    begin
                        e := individualData[i];
                        individualData[i] := individualData[k];
                        individualData[k] := e;
                        t := evolutionTotal[i];
                        evolutionTotal[i] := evolutionTotal[k];
                        evolutionTotal[k] := t;
                    end;
                end;
            end;
        end;

        procedure TLGP.Select(index: integer; value : double);
        begin
                evolutionTotal[index] := value;
        end;

        function TLGP.getEvolution(index : integer):double;
        begin
                result := evolutionTotal[index];
        end;

        procedure TLGP.Crossover(indexFrom, indexTo : integer);
        begin
                individualData[indexTo].setCrossover(individualData[indexFrom]);
        end;

        procedure TLGP.Mutation(index: integer);
        begin
                individualData[index].randomPopulate;
        end;

        procedure TLGP.setCrossoverRate(rate: double);
        begin
                if rate < 0 then
                    rate := 0;
                if rate > 1 then
                    rate := 1;
                crossoverRate := rate;
                crossoverPop := trunc(populationCount * crossoverRate);
        end;

        function TLGP.getCrossoverRate():double;
        begin
                result := crossoverRate;
        end;

        function TLGP.getIndividual(index:integer):TTuringEvaluator;
        begin
                result := individualData[index];
        end;

        procedure TLGP.setSample(s : TSampleData);
        begin
            if(sample <> nil) then
                    sample.Free;

            sample := s;

            Populate;
        end;

        function TLGP.getSample():TSampleData;
        begin
                result := sample;
        end;

end.
